/*
 * Copyright © 2022 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#define GOA_API_IS_SUBJECT_TO_CHANGE
#include <goa/goa.h>
#include "goamailclient.h"
#include "goamailutils.h"
#include "goasmtpauth.h"

#include <locale.h>
#include <gio/gunixinputstream.h>

#define SMTP_PORT_SSL   465
#define SMTP_PORT_TLS   587
#define SMTP_PORT_PLAIN 25

static gboolean
smtp_check_not_250 (const gchar *response, GError **error)
{
  if (!g_str_has_prefix (response, "250") || strlen (response) < 4)
    {
      g_set_error (error,
                   GOA_ERROR,
                   GOA_ERROR_FAILED, /* TODO: more specific */
                   "Unexpected response `%s'",
                   response);
      return TRUE;
    }

  return FALSE;
}

static gboolean
smtp_check_not_354 (const gchar *response, GError **error)
{
  if (!g_str_has_prefix (response, "354") || strlen (response) < 4)
    {
      g_set_error (error,
                   GOA_ERROR,
                   GOA_ERROR_FAILED, /* TODO: more specific */
                   "Unexpected response `%s'",
                   response);
      return TRUE;
    }

  return FALSE;
}

static gboolean
smtp_mail_from(GoaMailClient       *client,
               const char          *addr,
               GError             **error)
{
  gboolean ret = FALSE;
  GDataInputStream *input = goa_mail_client_get_input(client);
  GDataOutputStream *output = goa_mail_client_get_output(client);
  gchar *request = NULL;
  gchar *response = NULL;

  request = g_strdup_printf ("MAIL FROM:<%s>\r\n", addr);
  g_debug ("> %s", request);
  if (!g_data_output_stream_put_string (output, request, NULL, error))
    goto out;

  response = goa_utils_data_input_stream_read_line (input, NULL, NULL, error);
  if (response == NULL)
    goto out;
  g_debug ("< %s", response);
  if (smtp_check_not_250 (response, error))
    goto out;

  ret = TRUE;

out:
  g_free (request);
  g_free (response);
  return ret;
}

static gboolean
smtp_rcpt_to(GoaMailClient       *client,
             const char          *addr,
             GError             **error)
{
  gboolean ret = FALSE;
  GDataInputStream *input = goa_mail_client_get_input(client);
  GDataOutputStream *output = goa_mail_client_get_output(client);
  gchar *request = NULL;
  gchar *response = NULL;

  request = g_strdup_printf ("RCPT TO:<%s>\r\n", addr);
  g_debug ("> %s", request);
  if (!g_data_output_stream_put_string (output, request, NULL, error))
    goto out;

  response = goa_utils_data_input_stream_read_line (input, NULL, NULL, error);
  if (response == NULL)
    goto out;
  g_debug ("< %s", response);
  if (smtp_check_not_250 (response, error))
    goto out;

  ret = TRUE;

out:
  g_free (request);
  g_free (response);
  return ret;
}

#define IN_BUFFER_SIZE 16384
/* Repeated "\n." can be expanded to "\r\n.."  */
#define OUT_BUFFER_SIZE (IN_BUFFER_SIZE * 2)

enum {
  STATE_BOL,
  STATE_ANY,
  STATE_CR,
};

static gint
smtp_quote(gchar *outbuf, const gchar *buf, guint size, gint *p_state)
{
  guint out = 0;
  guint i = 0;
  gint state = *p_state;
  while (i < size)
    {
      switch (state)
        {
        case STATE_CR:
          if (buf[i] == '\n')
            state = STATE_BOL;
          else
            {
              state = STATE_ANY;
              goto any;
            }
          break;

        case STATE_BOL:
          if (buf[i] == '.')
            outbuf[out++] = '.';
          state = STATE_ANY;
          goto any;

        case STATE_ANY:
        any:
          if (buf[i] == '\n')
            {
              outbuf[out++] = '\r';
              state = STATE_BOL;
              break;
            }
          else
            {
              if (buf[i] == '\r')
                state = STATE_CR;
            }
          break;

	default:
	  g_assert_not_reached();
        }

      outbuf[out++] = buf[i++];
    }
  *p_state = state;
  return out;
}

static gboolean
smtp_send_body(GInputStream        *source,
               GOutputStream       *output,
               GError             **error)
{
  gboolean ret = FALSE;
  gchar *buf = g_malloc(IN_BUFFER_SIZE);
  gchar *outbuf = g_malloc(OUT_BUFFER_SIZE);
  gint state = STATE_BOL;

  for (;;)
    {
      gssize n = g_input_stream_read(source, buf, IN_BUFFER_SIZE, NULL, error);
      if (n <= 0)
        {
          ret = (n == 0);
          break;
        }

      n = smtp_quote(outbuf, buf, n, &state);
      g_debug ("> Sending %d bytes", (gint)n);
      if (g_output_stream_write(output, outbuf, n, NULL, error) < 0)
        break;
    }

  g_free (buf);
  g_free (outbuf);
  return ret;
}

static gboolean
smtp_data(GoaMailClient       *client,
          GInputStream        *source,
          GError             **error)
{
  gboolean ret = FALSE;
  GDataInputStream *input = goa_mail_client_get_input(client);
  GDataOutputStream *output = goa_mail_client_get_output(client);
  gchar *response = NULL;

  g_debug ("> DATA");
  if (!g_data_output_stream_put_string (output, "DATA\r\n", NULL, error))
    goto out;

  response = goa_utils_data_input_stream_read_line (input, NULL, NULL, error);
  if (response == NULL)
    goto out;
  g_debug ("< %s", response);
  if (smtp_check_not_354 (response, error))
    goto out;

  g_clear_pointer(&response, g_free);

  if (!smtp_send_body(source, G_OUTPUT_STREAM(output), error))
    goto out;

  g_debug ("> .");
  if (!g_data_output_stream_put_string (output, ".\r\n", NULL, error))
    goto out;

  response = goa_utils_data_input_stream_read_line (input, NULL, NULL, error);
  if (response == NULL)
    goto out;
  g_debug ("< %s", response);
  if (smtp_check_not_250 (response, error))
    goto out;

  ret = TRUE;

out:
  g_free (response);
  return ret;
}

static GoaMailClient *
get_goa_mail_client(const char *desired_sender)
{
  GError *error = NULL;
  GoaClient *client;
  GList *accounts, *l;
  guint i;

  gchar *sender = NULL;
  gchar *host_and_port = NULL;
  guint port = 0;
  GoaTlsType tls_type = GOA_TLS_TYPE_NONE;
  gchar *username = NULL;
  gchar *access_token = NULL;
  gchar *password = NULL;

  GoaMailAuth *mail_auth = NULL;
  GoaMailClient *mail_client = NULL;

  client = goa_client_new_sync (NULL, &error);
  if (!client) {
    g_error ("Could not create GoaClient: %s", error->message);
    exit(1);
  }

  accounts = goa_client_get_accounts (client);
  for (l = accounts, i = 0; l != NULL; l = l->next)
    {
      GoaAccount *account_obj = goa_object_get_account (GOA_OBJECT (l->data));
      GoaMail *mail_obj = goa_object_get_mail (GOA_OBJECT (l->data));
      GoaOAuth2Based *oauth2_obj = NULL;
      GoaPasswordBased *password_obj = NULL;

      if (!mail_obj)
        continue; 

      if (desired_sender)
        {
          if (!g_str_equal(desired_sender, goa_account_get_presentation_identity (account_obj)))
            continue;
          if (i == 1)
            g_printerr ("goa-sendmail: warning: More than one account defined for '%s'\n", desired_sender);
        }
      else
        {
          if (i > 0)
            {
              g_printerr ("goa-sendmail: More than one account defined, please use the -f option.\n");
              g_printerr ("goa-sendmail: If you are using git-send-email, you may want to execute\n");
              g_printerr ("goa-sendmail: \"git config git.sendemail.envelopesender auto\"\n");
              exit (1);
            }
        }

      host_and_port = g_strdup(goa_mail_get_smtp_host (mail_obj));
      if (goa_mail_get_smtp_use_ssl(mail_obj))
        {
          port = SMTP_PORT_SSL;
          tls_type = GOA_TLS_TYPE_SSL;
        }
      else if (goa_mail_get_smtp_use_tls(mail_obj))
        {
          port = SMTP_PORT_TLS;
          tls_type = GOA_TLS_TYPE_STARTTLS;
        }
      else
        {
          port = SMTP_PORT_PLAIN;
          tls_type = GOA_TLS_TYPE_NONE;
        }

      sender = g_strdup(goa_account_get_presentation_identity (account_obj));
      username = g_strdup(goa_mail_get_smtp_user_name (mail_obj));

      oauth2_obj = goa_object_get_oauth2_based (GOA_OBJECT (l->data));
      if (oauth2_obj && !goa_oauth2_based_call_get_access_token_sync (oauth2_obj,
                                                                      &access_token,
                                                                      NULL,
                                                                      NULL,
                                                                      NULL))
        {
          g_printerr("goa-sendmail: Could not retrieve OAuth2 access token\n");
          exit (1);
        }

      if (password_obj && !goa_password_based_call_get_password_sync (password_obj,
                                                                      "smtp-password",
                                                                      &password,
                                                                      NULL,
                                                                      NULL))
        {
          g_printerr("goa-sendmail: Could not retrieve password\n");
          exit (1);
        }

      g_clear_object (&mail_obj);
      g_clear_object (&oauth2_obj);
      g_clear_object (&password_obj);
      i++;
    }

  g_list_free_full (accounts, (GDestroyNotify) g_object_unref);
  g_object_unref(client);

  if (!host_and_port)
    {
      if (desired_sender)
        g_printerr("goa-sendmail: No mail accounts configured for '%s'\n", desired_sender);
      else
        g_printerr("goa-sendmail: No mail accounts configured\n");
      exit (1);
    }

  mail_auth = goa_smtp_auth_new_full("goa-sendmail", username, password, access_token);
  mail_client = goa_mail_client_new();
  if (!goa_mail_client_check_sync(mail_client,
                                  host_and_port, tls_type, FALSE, port,
                                  mail_auth, NULL, &error))
    {
      g_printerr("goa-sendmail: %s\n", error->message);
      exit(1);
    }

  if (!smtp_mail_from(mail_client, sender, &error))
    {
      g_printerr("goa-sendmail: %s\n", error->message);
      exit(1);
    }

  g_object_unref(mail_auth);
  return mail_client;
}


static gboolean until_eof;
static gchar *desired_sender;
static gchar **recipients;
static const GOptionEntry entries[] = {
  {
    "sender", 'f', 0, G_OPTION_ARG_STRING, &desired_sender,
    "The address from which mail will be sent",
    "ADDRESS"
  },
  {
    "read-until-eof", 'i', 0, G_OPTION_ARG_NONE, &until_eof,
    "This option is ignored",
    "ADDRESS"
  },
  {
    G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_STRING_ARRAY, &recipients,
    "The addresses to which mail will be sent",
    "ADDRESSES",
  },
  { NULL }
};

int
main (int argc, char **argv)
{
  GError *error = NULL;
  char **p_address;
  GoaMailClient *mail_client;
  GInputStream *source;
  GOptionContext *context;

  setlocale (LC_ALL, "");

  context = g_option_context_new("- sendmail interface to GNOME Online Accounts");
  g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
  if (!g_option_context_parse(context, &argc, &argv, &error))
    goto err;

  if (!recipients)
    {
      g_printerr("goa-sendmail: No recipients supplied\n");
      exit(0);
    }

  mail_client = get_goa_mail_client(desired_sender);
  for (p_address = recipients; *p_address; p_address++)
    if (!smtp_rcpt_to(mail_client, *p_address, &error))
      goto err;

  source = g_unix_input_stream_new(0, FALSE);
  if (!smtp_data(mail_client, source, &error))
    goto err;
  g_object_unref(source);

  g_object_unref(mail_client);
  return 0;

err:
  g_printerr("goa-sendmail: %s\n", error->message);
  exit(1);
}
